#ifndef DFSYS_H
#define DFSYS_H

#include "idxalloc.h"

struct varstrhashtable;

struct fsdir;

struct fsfile {
  struct fsdir* parentdir;
  unsigned int filename;
  unsigned int keyname;
  unsigned int fileidx; // low bit is file flag
};

struct fsdir {
  struct fsfile filedata;
  struct varstrhashtable* filenames;
  struct fsdir* dirs;
  struct fsfile* files;
  unsigned int filecap;
  unsigned int filenum;
  unsigned int dircap;
  unsigned int dirnum;
};

//

unsigned char fsinitdir(struct fsdir*, unsigned int dircap, unsigned int filecap, struct idxalloc*);

struct fsfile* fsaddfile(struct fsdir*, char const* name, char const* keyname, struct idxalloc*);
struct fsdir* fsadddirectory(struct fsdir*, char const* name, char const* keyname, struct idxalloc*);

void fsfreedir(struct fsdir* dir);

// remove file from parent
// if directory, also removes all files and subdirectories, then frees structs
// then returns file id
unsigned int fsremovefile(struct fsfile*, struct idxalloc*);

char const* fsgetfilename(struct fsfile const*);
char const* fsgetkeyname(struct fsfile const*);

void fsgetfilepath(struct fsfile const*, char*);
void fsgetdirpath(struct fsdir const*, char*);


struct fsfile* fsgetfile(struct fsdir* dir, char const* keyname);

// resolve by keynamed
struct fsfile* fsresolvepath(struct fsdir*, char const* path);

#endif //DFSYS_H