#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MINFATCLUSTER 2

char nextline(FILE* file, char* buffer, unsigned int buffersize, int* buffered, int* lineofs, int* lineend, char* state) {
  int nstate = *state;
  char eos = 0;
  int blen = *buffered;
  int ofs = *lineend;
  if ((nstate & 1)) {
    if (ofs < blen) {
      ofs += (buffer[ofs] == '\n');
    } else {
      ofs = 0;
      blen = fread(buffer, 1, buffersize, file);
      *buffered = blen;
      if (ofs < blen) {
        ofs += (buffer[ofs] == '\n');
      }
    }
  }
  nstate = 0;
  int end = ofs;
  char foundend = 0;
  do {
    if (end == blen) {
      if ((unsigned int)blen == buffersize) {
        if (ofs > 0) {
          memmove(buffer, buffer + ofs, end - ofs);
          end -= ofs;
          blen -= ofs;
          ofs = 0;
        }
        if ((unsigned int)blen < buffersize) {
          blen += fread(buffer + blen, 1, buffersize - blen, file);
          *buffered = blen;
        }
      } else {
        eos = 1;
      }
    }
    foundend = 1;
    while (end < blen && !(foundend = ((buffer[end] == '\n') | (buffer[end] == '\r')))) {
      ++end;
    }
    if (end < blen) {
      nstate = (buffer[end] == '\r');
      ++end;
    }
  } while (!foundend);
  *state = nstate;
  *lineofs = ofs;
  *lineend = end;
  return !eos;
}

void bufferwrite(FILE* file, char* writebuffer, int buffersize, char const* data, int len, int* buffered) {
  int blen = *buffered;
  if (blen + len <= buffersize) {
    memcpy(writebuffer + blen, data, len);
    *buffered = blen + len;
  } else {
    while (len > 0) {
      int rem = buffersize - blen;
      int clen = len > rem ? rem : len;
      memcpy(writebuffer + blen, data, clen);
      blen += clen;
      fwrite(writebuffer, 1, blen, file);
      blen = 0; //
      data += clen;
      len -= clen;
    }
    *buffered = blen;
  }
}

void bufferwritez(FILE* file, char* writebuffer, int buffersize, char const* str, int* buffered) {
  bufferwrite(file, writebuffer, buffersize, str, strlen(str), buffered);
}

int getu16l(char const* buffer) {
  return ((int)buffer[0] & 0xff) | ((int)buffer[1] & 0xff) << 8;
}

int getu24l(char const* buffer) {
  return ((int)buffer[0] & 0xff) | ((int)buffer[1] & 0xff) << 8 | ((int)buffer[2] & 0xff) << 16;
}

int geti32l(char const* buffer) {
  return ((int)buffer[0] & 0xff) | ((int)buffer[1] & 0xff) << 8 | ((int)buffer[2] & 0xff) << 16 | ((int)buffer[3] & 0xff) << 24;
}

struct diskprops {
  int prediskheader;
  int nextcluster[6144];
  char* dircache;
  int cachedirs;
  int* dirmeta; // first sector, start cache ofs, end cache ofs
  int dircachecap;
  int dircachesize;
  int dirmetacap;

  int sectorbytes;
  int sectorspercluster;
  int reservedsectors;
  int numfats;
  int maxrootdirentries;
  int numsectors16;
  int mediadesc;
  int sectorsperfat;

  int rootsectors;
  int fatsector;
  int rootsector;
  int datasector;
  int clustersize;
};

char ensuredirspace(struct diskprops* prop, int bytes, int entries) {
  char success = 1;
  if (prop->dircachesize + bytes > prop->dircachecap) {
    int newcap = prop->dircachecap;
    newcap = (newcap == 0) ? 1024 : 0;
    while (prop->dircachesize + bytes > newcap) {
      newcap *= 2;
    }
    char* newarray = realloc(prop->dircache, newcap);
    if (newarray != NULL) {
      prop->dircache = newarray;
      prop->dircachecap = newcap;
    }
  }
  if (prop->cachedirs + entries > prop->dirmetacap) {
    int newcap = prop->dirmetacap;
    newcap = (newcap == 0) ? 16 : 0;
    while (prop->cachedirs + entries > newcap) {
      newcap *= 2;
    }
    int* newarray = realloc(prop->dirmeta, sizeof(int) * (newcap * 3));
    if (newarray != NULL) {
      prop->dirmeta = newarray;
      prop->dirmetacap = newcap;
    }
  }
  return success;
}

int adddirectoryentry(FILE* disk, struct diskprops* prop, int firstcluster) {
  int index = -1;
  int const dataofs = prop->prediskheader + prop->datasector * prop->sectorbytes;
  int lastcluster = firstcluster;

  char success = 1;
  char ctd = 1;
  
  int cachestart = prop->dircachesize;
  int cacheofs = cachestart;

  int nextcluster = firstcluster;
  while (ctd) {
    success = ensuredirspace(prop, cacheofs - cachestart + prop->clustersize, 0);
    if (success) {
      if (nextcluster != lastcluster + 1) {
        success = (fseek(disk, dataofs + (nextcluster - (MINFATCLUSTER)) * prop->clustersize, SEEK_SET) == 0);
      }
      int readamt = success ? fread(prop->dircache + cacheofs, 1, prop->clustersize, disk) : 0;
      if (readamt == prop->clustersize) {
        cacheofs += readamt;
      } else {
        success = 0;
      }
      lastcluster = nextcluster;
      nextcluster = prop->nextcluster[nextcluster];
      if (nextcluster >= 0xff0) {
        ctd = 0;
      }
    }
    ctd &= success;
  }
  if (success) {
    prop->dirmeta[prop->cachedirs * 3] = firstcluster;
    prop->dirmeta[prop->cachedirs * 3 + 1] = cachestart;
    prop->dirmeta[prop->cachedirs * 3 + 2] = cacheofs;
    index = prop->cachedirs++;
    prop->dircachesize = cacheofs;
  }
  return index;
}

// get index in cache list for directory info, -1 if not found
int getdirectoryentry(FILE* disk, struct diskprops* prop, int firstcluster) {
  int index = -1;
  for (int i = 0; index == -1 && i < prop->cachedirs; ++i) {
    if (prop->dirmeta[i * 3] == firstcluster) {
      index = i;
    }
  }
  if (index == -1) { // try to read to cache
    index = adddirectoryentry(disk, prop, firstcluster);
  }
  return index;
}

int initrootdir(FILE* disk, struct diskprops* prop) {
  int const toread = prop->rootsectors * prop->sectorbytes;
  char success = ensuredirspace(prop, toread, 1);
  if (success) {
    success = (fseek(disk, prop->prediskheader + prop->rootsector * prop->sectorbytes, SEEK_SET) == 0);
  }
  int idx = -1;
  if (success) {
    int readamt = fread(prop->dircache, 1, toread, disk);
    if (readamt == toread) {
      prop->dirmeta[prop->cachedirs * 3] = -1;
      prop->dirmeta[prop->cachedirs * 3 + 1] = prop->dircachesize;
      prop->dirmeta[prop->cachedirs * 3 + 2] = prop->dircachesize + toread;
      idx = prop->cachedirs++;
      prop->dircachesize += toread;
    }
  }
  return idx;
}

char nameequals(char const* name831, char const* name832) {
  char matches = 0;
  int ofs = 0;
  while (ofs < 8 && name831[ofs] != ' ' && name831[ofs] == name832[ofs]) {
    ++ofs;
  }
  matches = (ofs == 8 || ((name831[ofs] == ' ') & (name832[ofs] == ' ')));
  if (matches) {
    ofs = 8;
    while (ofs < 11 && name831[ofs] != 0 && name831[ofs] == name832[ofs]) {
      ++ofs;
    }
    matches = (ofs == 11 || ((name831[ofs] == ' ') & (name832[ofs] == ' ')));
  }
  return matches;
}

int getentryindex(char const* buffer, int start, int end, char const* name83, int withflags, int withoutflags) {
  char found = 0;
  int dirofs = start;
  int foundidx = -1;
  while (!found && dirofs + 32 <= end) {
    ++foundidx;
    int flags = (int)buffer[dirofs + 11];
    if ((flags & withflags) == withflags && (flags & withoutflags) == 0) {
      if (nameequals(buffer + dirofs, name83)) {
        found = 1;
      }
    }
    dirofs += 32;
  }
  return found ? foundidx : -1;
}

void nametext(char const* src, char* dst, int nlen) {
  char prevdb = 0;
  for (int i = 0; i < nlen; ++i) {
    char c = src[i];
    if (!prevdb) {
      int temp = (int)c & 0xff;
      prevdb = ((temp >= 0x81) & (temp < 0xa0)) | ((temp >= 0xe0) & (temp < 0xfd));
      dst[i] = c - ((c >= 'a') & (c <= 'z')) * 32;
    } else {
      dst[i] = c;
      prevdb = 0;
    }
  }
}

void nameto83(char const* src, int srcstart, int srcend, char* dst) {
  for (int i = 0; i < 11; ++i) {
    dst[i] = ' ';
  }
  if (srcend > srcstart) {
    int dot = srcend - 1;
    while (dot > srcstart && src[dot] != '.') {
      --dot;
    }
    if (dot != srcstart) { // no dot
      int ext = dot + 1;
      int nlen = dot - srcstart;
      if (nlen >= 8) {
        nlen = 8;
      }
      nametext(src, dst, nlen);
      int extlen = srcend - ext;
      if (extlen >= 3) {
        extlen = 3;
      }
      nametext(src + ext, dst + 8, extlen);
    } else if (src[dot] != '.') {
      int nlen = srcend - srcstart;
      if (nlen >= 11) {
        nlen = 11;
      }
      nametext(src, dst, nlen);
    }
  }
}

int findbyte(char const* str, int start, int end, char search) {
  char prevdb = 0;
  int foundofs = -1;
  int ofs = start;
  while (foundofs == -1 && ofs < end) {
    char c = str[ofs];
    if (!prevdb) {
      int temp = (int)c & 0xff;
      prevdb = ((temp >= 0x81) & (temp < 0xa0)) | ((temp >= 0xe0) & (temp < 0xfd));
      if (c == search) {
        foundofs = ofs;
      }
    } else {
      prevdb = 0;
    }
    ++ofs;
  }
  return foundofs;
}

int nextslash(char const* str, int start, int end) {
  return findbyte(str, start, end, '\\');
}

int getfilecluster(FILE* disk, struct diskprops* prop, char const* fpath, int pathlen) {
  int const FLAG_DIRECTORY = 0x10;
  int const FLAG_VOLUME = 0x08;

  int cluster = -1;
  char name83[11];
  int slashofs = -1;
  int diridx = 0;
  int pathofs = 0;
  do {
    slashofs = nextslash(fpath, pathofs, pathlen);
    if (slashofs != -1) {
      nameto83(fpath, pathofs, slashofs, name83);
      int entryidx = getentryindex(prop->dircache, prop->dirmeta[diridx * 3 + 1], prop->dirmeta[diridx * 3 + 2], name83, FLAG_DIRECTORY, FLAG_VOLUME);
      if (entryidx != -1) {
        int firstcluster = getu16l(prop->dircache + (prop->dirmeta[diridx * 3 + 1] + entryidx * 32 + 26));
        diridx = getdirectoryentry(disk, prop, firstcluster);
      } else {
        diridx = -1;
      }
      pathofs = slashofs + 1;
    }
  } while (diridx != -1 && (slashofs != -1 || slashofs + 1 == pathlen));
  if (diridx != -1) {
    if (slashofs == -1) {
      nameto83(fpath, pathofs, pathlen, name83);
      int entryidx = getentryindex(prop->dircache, prop->dirmeta[diridx * 3 + 1], prop->dirmeta[diridx * 3 + 2], name83, 0, FLAG_VOLUME);
      if (entryidx != -1) {
        cluster = getu16l(prop->dircache + (prop->dirmeta[diridx * 3 + 1] + entryidx * 32 + 26));
      }
    } else {
      cluster = prop->dirmeta[diridx * 3 + 0];
    }
  }
  return cluster;
}

// get the sector (not cluster) a file's data starts on (if the file is fragmented, this won't be useful by itself)
int getfilesector(FILE* disk, struct diskprops* prop, char const* fpath, int pathlen) {
  int cluster = getfilecluster(disk, prop, fpath, pathlen);
  return cluster != -1 ? (cluster - (MINFATCLUSTER)) * prop->sectorspercluster + prop->datasector : -1;
}

int getlineitems(char const* buffer, int start, int end, int* ofspairs, int maxpairs) {
  int ofs = start;
  int numpairs = 0;
  while (ofs < end && numpairs + 1 < maxpairs) {
    while (ofs < end && buffer[ofs] == ' ') {
      ++ofs;
    }
    int ofs0 = ofs;
    while (ofs < end && buffer[ofs] != ' ') {
      ++ofs;
    }
    if (ofs < end) {
      ofspairs[numpairs * 2] = ofs0;
      ofspairs[numpairs * 2 + 1] = ofs;
      ++numpairs;
    }
  }
  while (ofs < end && buffer[ofs] == ' ') {
    ++ofs;
  }
  while (end > ofs && buffer[end - 1] == ' ') {
    --end;
  }
  if (ofs < end) {
    ofspairs[numpairs * 2] = ofs;
    ofspairs[numpairs * 2 + 1] = end;
    ++numpairs;
  }
  return numpairs;
}

int printnum(char* buffer, int val) {
  char neg = (val < 0);
  if (neg) {
    val = -val;
  }
  int len = 0;
  buffer[len++] = '0' + (val % 10);
  val /= 10;
  while (val > 0) {
    buffer[len++] = '0' + (val % 10);
    val /= 10;
  }
  if (neg) {
    buffer[len++] = '-';
  }
  buffer[len] = 0;
  int hlen = len / 2;
  for (int i = 0; i < hlen; ++i) {
    int eofs = len - i - 1;
    char temp = buffer[i];
    buffer[i] = buffer[eofs];
    buffer[eofs] = temp;
  } 
  return len;
}

int main(int argc, char const** argv) {
  char success = 1;

  if (argc == 4) {
    FILE* outfile = fopen(argv[3], "wb");
    if (outfile != NULL) {
      int const readbuffersize = 16384;
      char readbuffer[readbuffersize];
      
      struct diskprops dprops;

      dprops.dircache = malloc(16384);
      dprops.cachedirs = 0;
      dprops.dirmeta = dprops.dircache != NULL ? malloc(sizeof(int) * (32 * 3)) : NULL;
      dprops.dircachecap = dprops.dircache != NULL ? 16384 : 0;
      dprops.dircachesize = 0;
      dprops.dirmetacap = dprops.dirmeta != NULL ? 32 : 0;

      int const writebuffersize = 16384;
      char writebuffer[writebuffersize];

      char tempnum[16];
      int writebuffered = 0;
      int readbuffered = 0;
      int lineofs = 0;
      int lineend = 0;
      char state = 0;
      char hasmore = 1;

      int lineofspairs[4];

      FILE* diskfile = fopen(argv[1], "rb");
      if (diskfile != NULL) {
        FILE* listfile = fopen(argv[2], "rb");
        if (listfile != NULL) {
          dprops.prediskheader = 162;
          readbuffered = fread(readbuffer, 1, dprops.prediskheader + 1024, diskfile);

          if (readbuffered < dprops.prediskheader + 128) {
            printf("Could not read disk properties\n");
            success = 0;
          }

          if (success) {
            int diskimdataofs = dprops.prediskheader;

            dprops.sectorbytes = getu16l(readbuffer + (diskimdataofs + 11));
            dprops.sectorspercluster = (int)readbuffer[diskimdataofs + 13] & 0xff;
            dprops.reservedsectors = getu16l(readbuffer + (diskimdataofs + 14));
            dprops.numfats = (int)readbuffer[diskimdataofs + 16] & 0xff;
            dprops.maxrootdirentries = getu16l(readbuffer + (diskimdataofs + 17));
            dprops.numsectors16 = getu16l(readbuffer +(diskimdataofs + 19));
            dprops.mediadesc = (int)readbuffer[diskimdataofs + 21] & 0xff;
            dprops.sectorsperfat = getu16l(readbuffer + (diskimdataofs + 22));

            int numclusterpairs = (dprops.sectorsperfat * dprops.sectorbytes * 8) / 24;
            int fatofs = dprops.reservedsectors * dprops.sectorbytes + dprops.prediskheader;

            if (numclusterpairs > 3072) { //
              numclusterpairs = 3072;
            }

            if (readbuffered != fatofs) {
              success = (fseek(diskfile, fatofs, SEEK_SET) == 0);
            }

            readbuffered = success ? fread(readbuffer, 1, dprops.numfats * dprops.sectorbytes, diskfile) : 0;
            if (readbuffered < dprops.numfats * dprops.sectorbytes) {
              printf("Could not read FAT table\n");
              success = 0;
            }
          
            if (success) {
              for (int i = 0; i < numclusterpairs; ++i) {
                int cpair = getu24l(readbuffer + (i * 3));
                dprops.nextcluster[i * 2] = cpair & 0x0fff;
                dprops.nextcluster[i * 2 + 1] = cpair >> 12 & 0x0fff;
              }
            }
          }

          if (success) {
            dprops.rootsectors = (dprops.maxrootdirentries * 32 + dprops.sectorbytes - 1) / dprops.sectorbytes; // not +32
            
            dprops.fatsector = dprops.reservedsectors;
            dprops.rootsector = dprops.fatsector + dprops.sectorsperfat * dprops.numfats;
            dprops.datasector = dprops.rootsector + dprops.rootsectors;
            dprops.clustersize = dprops.sectorspercluster * dprops.sectorbytes;

            int res = initrootdir(diskfile, &dprops);
            if (res == -1) {
              printf("Could not read root directory of disk\n");
              success = 0;
            }
          }
          if (success) {
            // process list file
            int linenum = 0;
            readbuffered = fread(readbuffer, 1, readbuffersize, listfile);
            do {
              ++linenum;
              hasmore = nextline(listfile, readbuffer, readbuffersize, &readbuffered, &lineofs, &lineend, &state);
              int len2 = lineend - lineofs;
              int eofs = lineend;
              if (len2 >= 1 && ((readbuffer[lineend - 1] == '\r') | (readbuffer[lineend - 1] == '\n'))) {
                --eofs;
              }
              if (len2 >= 2 && ((readbuffer[lineend - 2] == '\r'))) {
                --eofs;
              }
              int numvals = getlineitems(readbuffer, lineofs, eofs, lineofspairs, 2);
              if (numvals == 2) {
                int sec = getfilesector(diskfile, &dprops, readbuffer + lineofspairs[2], lineofspairs[3] - lineofspairs[2]);
                if (sec != -1) {
                  bufferwritez(outfile, writebuffer, writebuffersize, "%define ", &writebuffered);
                  bufferwrite(outfile, writebuffer, writebuffersize, readbuffer + lineofspairs[0], lineofspairs[1] - lineofspairs[0], &writebuffered);
                  bufferwritez(outfile, writebuffer, writebuffersize, " ", &writebuffered);
                  int numlen = printnum(tempnum, sec);
                  bufferwrite(outfile, writebuffer, writebuffersize, tempnum, numlen, &writebuffered);
                  bufferwritez(outfile, writebuffer, writebuffersize, "\n", &writebuffered);
                }
              }
            } while (success && hasmore);
            if (writebuffered > 0) {
              fwrite(writebuffer, 1, writebuffered, outfile);
            }
          }
        } else{ 
          printf("Could not open sector fetch list file %s for reading\n", argv[2]);
          success = 0;
        }
        fclose(diskfile);
      } else {
        printf("Could not open DCP file %s for reading\n", argv[1]);
        success = 0;
      }
      free(dprops.dircache);
      free(dprops.dirmeta);
    } else {
      printf("Could not open output file %s for writing\n", argv[3]);
        success = 0;
    }
  } else {
    printf("Expects three arguments.");
    success = 0;
  }
  fflush(stdout);
  return !success;
}