#include "dfsys.h"

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

unsigned int vshashstr(char const* str, unsigned int len) {
  unsigned int hval = 0;
  for (unsigned int idx = 0; idx < len; ++idx) {
    hval = (hval * 31 + ((unsigned int)str[idx] & 0xff));
  }
  return hval & 0xffffffff;
}

// removable entries
struct varstrhashtable {
  char* strdata;
  unsigned int* entries; // hash, ofs, length, nextentry
  unsigned int* bins;
  unsigned int entrynum;
  unsigned int entrycap;
  unsigned int dataofs;
  unsigned int datasize;
  unsigned int datacap;
  unsigned int lastunalloc;
};

// return idx or (unsigned int)-1 if can't add
unsigned int vsaddstr(struct varstrhashtable* stable, char const* str, unsigned int slen, unsigned int vidx) {
  unsigned int foundidx = (unsigned int)-1;
  unsigned int hadd = vshashstr(str, slen);
  unsigned int sfirst = stable->bins[hadd % stable->entrycap],
               shecond = (unsigned int)-1;
  while (sfirst != (unsigned int)-1 && (slen != (stable->entries[sfirst * 4 + 2] & 0xffff) || hadd != stable->entries[sfirst * 4] || memcmp(str, stable->strdata + stable->entries[sfirst * 4 + 1], slen))) {
    shecond = sfirst;
    sfirst = stable->entries[sfirst * 4 + 3];
  }
  if (sfirst == (unsigned int)-1) { // add new
    unsigned char result = 1;
    if (stable->dataofs + slen + 1 > stable->datacap) {
      unsigned int newcap = stable->datacap == 0 ? 128 : stable->datacap;
      while ((stable->datasize + slen + 1) * 2 > newcap) {
        newcap *= 2;
      }
      char* newdata = (char*)malloc(sizeof(char) * newcap);
      if (newdata != NULL) {
        unsigned int dstofs = 0;
        for (unsigned int idx = 0; idx < stable->entrycap; ++idx) {
          if (stable->entries[idx * 4 + 2] != (unsigned int)-1) {
            memcpy(newdata + dstofs, stable->strdata + stable->entries[idx * 4 + 1], stable->entries[idx * 4 + 2] & 0xffff);
            stable->entries[idx * 4 + 1] = dstofs;
            dstofs += stable->entries[idx * 4 + 2] & 0xffff;
            newdata[dstofs++] = 0;
          }
        }
        stable->dataofs = dstofs;
        stable->datasize = dstofs;
        free(stable->strdata);
        stable->strdata = newdata;
        stable->datacap = newcap;
      } else {
        result = 0;
      }
    }
    if (result && stable->entrynum + 1 == stable->entrycap) {
      unsigned int* newarr = (unsigned int*)realloc(stable->entries, sizeof(unsigned int) * stable->entrycap * 4 * 2);
      if (newarr != NULL) {
        stable->entries = newarr;
      } else {
        result = 0;
      }
      newarr = newarr != NULL ? (unsigned int*)realloc(stable->bins, sizeof(unsigned int) * stable->entrycap * 2) : NULL;
      if (newarr != NULL) {
        stable->bins = newarr;
        memset(stable->bins, -1, sizeof(unsigned int) * stable->entrycap * 2);
        for (unsigned int idx = stable->entrycap; idx + 1 < stable->entrycap * 2; ++idx) {
          stable->entries[idx * 4 + 2] = (unsigned int)-1;
          stable->entries[idx * 4 + 3] = (idx + 1);
        }
        stable->lastunalloc = stable->entrycap * 2 - 1;
        stable->entries[stable->lastunalloc * 4 + 2] = (unsigned int)-1;
        stable->entries[stable->lastunalloc * 4 + 3] = stable->entrycap;
        stable->entrycap = stable->entrycap * 2;
      } else {
        result = 0;
      }
      if (result) { // rehash
        for (unsigned int idx = 0; idx < stable->entrynum; ++idx) {
          unsigned int binidx = stable->entries[idx * 4] % stable->entrycap;
          unsigned int lastidx = stable->bins[binidx];
          if (lastidx == (unsigned int)-1) {
            stable->entries[idx * 4 + 3] = idx;
          } else {
            stable->entries[idx * 4 + 3] = stable->entries[lastidx * 4 + 3]; // copy first
            stable->entries[lastidx * 4 + 3] = idx;
          }
          stable->bins[binidx] = idx;
        }
        shecond = stable->bins[hadd % stable->entrycap];
        for (unsigned int idx = 0; idx < stable->entrycap; ++idx) { // split
          unsigned int lastidx = stable->bins[idx];
          if (lastidx != (unsigned int)-1) {
            stable->bins[idx] = stable->entries[lastidx * 4 + 3];
            stable->entries[lastidx * 4 + 3] = (unsigned int)-1;
          }
        }
      }
    }
    if (result) {
      foundidx = stable->entries[stable->lastunalloc * 4 + 3];
      stable->entries[stable->lastunalloc * 4 + 3] = stable->entries[foundidx * 4 + 3];
      memcpy(stable->strdata + stable->dataofs, str, slen);
      stable->strdata[stable->dataofs + slen] = 0;
      if (shecond != (unsigned int)-1) {
        stable->entries[shecond * 4 + 3] = foundidx;
      } else {
        stable->bins[hadd % stable->entrycap] = foundidx;
      }
      stable->entries[foundidx * 4] = hadd;
      stable->entries[foundidx * 4 + 1] = stable->dataofs;
      stable->entries[foundidx * 4 + 2] = vidx << 16 | (slen & 0xffff);
      stable->entries[foundidx * 4 + 3] = (unsigned int)-1;
      ++stable->entrynum;
      stable->datasize += slen + 1;
      stable->dataofs += slen + 1;
    }
  }
  return foundidx;
}

unsigned int vsfindstr(struct varstrhashtable const* stable, char const* str, unsigned int slen) {
  unsigned int hadd = vshashstr(str, slen);
  unsigned int sfirst = stable->bins[hadd % stable->entrycap];
  while (sfirst != (unsigned int)-1 && (slen != (stable->entries[sfirst * 4 + 2] & 0xffff) || hadd != stable->entries[sfirst * 4] || memcmp(str, stable->strdata + stable->entries[sfirst * 4 + 1], slen))) {
    sfirst = stable->entries[sfirst * 4 + 3];
  }
  return sfirst;
}

// return vidx
unsigned int vsremovebyidx(struct varstrhashtable* stable, unsigned int eidx) {
  unsigned int hadd = stable->entries[eidx * 4];
  unsigned int sfirst = stable->bins[hadd % stable->entrycap],
               shecond = (unsigned int)-1;
  unsigned int vidx = (unsigned int)-1;
  while (sfirst != (unsigned int)-1 && sfirst != eidx) {
    shecond = sfirst;
    sfirst = stable->entries[sfirst * 4 + 3];
  }
  if (sfirst == eidx) {
    --stable->entrynum;
    if (shecond != (unsigned int)-1) {
      stable->entries[shecond * 4 + 3] = stable->entries[sfirst * 4 + 3];
    } else {
      stable->bins[hadd % stable->entrycap] = stable->entries[sfirst * 4 + 3];
    }
    stable->entries[sfirst * 4 + 3] = stable->entries[stable->lastunalloc * 4 + 3];
    stable->entries[stable->lastunalloc * 4 + 3] = sfirst;
    stable->lastunalloc = sfirst;
    vidx = stable->entries[sfirst * 4 + 2] >> 16;
    stable->datasize -= (stable->entries[sfirst * 4 + 2] & 0xffff) + 1;
    stable->entries[sfirst * 4 + 2] = (unsigned int)-1;
  }
  return vidx;
}

// return vidx of removed entry, or -1 if not found
unsigned int vsremovestr(struct varstrhashtable* stable, char const* str, unsigned int slen) {
  unsigned int hadd = vshashstr(str, slen);
  unsigned int sfirst = stable->bins[hadd % stable->entrycap],
               shecond = (unsigned int)-1;
  unsigned int vidx = (unsigned int)-1;
  while (sfirst != (unsigned int)-1 && (slen != (stable->entries[sfirst * 4 + 2] & 0xffff) || hadd != stable->entries[sfirst * 4] || memcmp(str, stable->strdata + stable->entries[sfirst * 4 + 1], slen))) {
    shecond = sfirst;
    sfirst = stable->entries[sfirst * 4 + 3];
  }
  if (sfirst != (unsigned int)-1) { // found
    --stable->entrynum;
    if (shecond != (unsigned int)-1) {
      stable->entries[shecond * 4 + 3] = stable->entries[sfirst * 4 + 3];
    } else {
      stable->bins[hadd % stable->entrycap] = stable->entries[sfirst * 4 + 3];
    }
    stable->entries[sfirst * 4 + 3] = stable->entries[stable->lastunalloc * 4 + 3];
    stable->entries[stable->lastunalloc * 4 + 3] = sfirst;
    stable->lastunalloc = sfirst;
    vidx = stable->entries[sfirst * 4 + 2] >> 16;
    stable->datasize -= (stable->entries[sfirst * 4 + 2] & 0xffff) + 1;
    stable->entries[sfirst * 4 + 2] = (unsigned int)-1;
  }
  return vidx;
}

void vshashtablefree(struct varstrhashtable* stable) {
  free(stable->bins);
  free(stable->entries);
  free(stable->strdata);
  stable->bins = NULL;
  stable->entries = NULL;
  stable->strdata = NULL;
  stable->entrynum = 0;
  stable->entrycap = 0;
  stable->datasize = 0;
  stable->datacap = 0;
  stable->lastunalloc = (unsigned int)-1;
}

void vshashtableclear(struct varstrhashtable* stable) {
  for (unsigned int idx = 0; idx < stable->entrycap; ++idx) {
    stable->entries[idx * 4 + 2] = (unsigned int)-1;
    stable->entries[idx * 4 + 3] = (idx + 1) % stable->entrycap;
  }
  stable->entrynum = 0;
  stable->datasize = 0;
  stable->dataofs = 0;
  stable->lastunalloc = stable->entrycap - 1;
  memset(stable->bins, -1, sizeof(unsigned int) * stable->entrycap);
}

unsigned char vshashtableinit(struct varstrhashtable* stable, unsigned int datacap, unsigned int entrycap) {
  unsigned char result = 1;
  stable->strdata = malloc(sizeof(char) * datacap);
  stable->entries = stable->strdata == NULL ? NULL : malloc(sizeof(unsigned int) * entrycap * 4);
  stable->bins = stable->entries == NULL ? NULL : malloc(sizeof(unsigned int) * entrycap);
  if (stable->bins != NULL)  {
    stable->entrycap = entrycap;
    stable->datacap = datacap;
    vshashtableclear(stable);
  } else {
    vshashtablefree(stable);
    result = 0;
  }
  return result;
}

unsigned char fsinitdir(struct fsdir* dir, unsigned int dircap, unsigned int filecap, struct idxalloc* ial) {
  unsigned char result = 1;
  dir->dirs = malloc(sizeof(struct fsdir) * dircap);
  dir->files = dir->dirs != NULL ? malloc(sizeof(struct fsfile) * filecap) : NULL;
  dir->filenames = dir->files != NULL ? malloc(sizeof(struct varstrhashtable)) : NULL;
  if (dir->filenames != NULL) {
    result = vshashtableinit(dir->filenames, 32 * (dircap + filecap), 2 * (dircap + filecap));
  } else {
    result = 0;
  }
  if (result) {
    dir->filecap = filecap;
    dir->dircap = dircap;
    dir->filenum = 0;
    dir->dirnum = 0;
    for (unsigned int idx = 0; idx < dir->filecap; ++idx) {
      struct fsfile* file = &dir->files[idx];
      file->parentdir = dir;
      file->keyname = (unsigned int)-1;
    }
    for (unsigned int idx = 0; idx < dir->dircap; ++idx) {
      struct fsdir* subdir = &dir->dirs[idx];
      struct fsfile* fdat = &subdir->filedata;
      fdat->parentdir = dir;
      fdat->keyname = (unsigned int)-1;
      subdir->dirs = NULL;
    }
    struct fsfile* fdat = &dir->filedata;
    fdat->parentdir = NULL;
    fdat->fileidx = idxgetnew(ial) << 1 | 0;
  } else {
    free(dir->filenames);
    free(dir->files);
    free(dir->dirs);
    dir->filenames = NULL;
    dir->files = NULL;
    dir->dirs = NULL;
  }
  return result;
}

unsigned int fsgetnewfile(struct fsdir* dir) {
  unsigned int fidx = (unsigned int)-1;
  if (dir->filenum == dir->filecap) {
    struct fsfile* newarr = (struct fsfile*)realloc(dir->files, sizeof(struct fsfile) * dir->filecap * 2);
    if (newarr != NULL) {
      dir->files = newarr;
      for (unsigned int cidx = 0; cidx < dir->filecap * 2; ++cidx) {
        struct fsfile* file = &dir->files[cidx];
        file->parentdir = dir;
      }
      for (unsigned int cidx = dir->filecap; cidx < dir->filecap * 2; ++cidx) {
        struct fsfile* file = &dir->files[cidx];
        file->keyname = (unsigned int)-1;
      }
      dir->filecap = dir->filecap * 2;
    }
  }
  if (dir->filenum < dir->filecap) {
    for (unsigned int idx = 0; fidx == (unsigned int)-1 && idx < dir->filecap; ++idx) {
      if (dir->files[idx].keyname == (unsigned int)-1) {
        fidx = idx;
      }
    }
  }
  return fidx;
}

unsigned int fsgetnewdir(struct fsdir* dir) {
  unsigned int fidx = (unsigned int)-1;
  if (dir->dirnum == dir->dircap) {
    struct fsdir* newarr = (struct fsdir*)realloc(dir->dirs, sizeof(struct fsdir) * dir->dircap * 2);
    if (newarr != NULL) {
      dir->dirs = newarr;
      for (unsigned int cidx = 0; cidx < dir->dircap * 2; ++cidx) {
        struct fsfile* file = &dir->dirs[cidx].filedata;
        file->parentdir = dir;
      }
      for (unsigned int cidx = dir->dircap; cidx < dir->dircap * 2; ++cidx) {
        struct fsdir* dir2 = &dir->dirs[cidx];
        dir2->dirs = NULL;
        struct fsfile* file = &dir2->filedata;
        file->keyname = (unsigned int)-1;
      }
      dir->dircap = dir->dircap * 2;
    }
  }
  if (dir->dirnum < dir->dircap) {
    for (unsigned int idx = 0; fidx == (unsigned int)-1 && idx < dir->dircap; ++idx) {
      if (dir->dirs[idx].filedata.keyname == (unsigned int)-1) {
        fidx = idx;
      }
    }
  }
  return fidx;
}

struct fsfile* fsaddfile(struct fsdir* dir, char const* keyname, char const* name, struct idxalloc* ial) {
  struct varstrhashtable* vsh = dir->filenames;
  unsigned int eidx1 = vsaddstr(vsh, keyname, strlen(keyname), 0);
  struct fsfile* newfile = NULL;
  if (eidx1 != (unsigned int)-1) {
    unsigned int fidx = fsgetnewfile(dir);
    if (fidx != (unsigned int)-1) {
      vsh->entries[eidx1 * 4 + 2] |= (fidx << 1 | 1) << 16;
      newfile = &dir->files[fidx];
      unsigned int eidx2 = vsaddstr(vsh, name, strlen(name), 0);
      newfile->keyname = eidx1;
      newfile->filename = eidx2 != (unsigned int)-1 ? eidx2 : eidx1;
      newfile->fileidx = idxgetnew(ial) << 1 | 1;
      if (eidx2 != (unsigned int)-1) {
        vsh->entries[eidx2 * 4 + 2] |= (fidx << 1 | 1) << 16;
      }
      ++dir->filenum;
    }
  }
  return newfile;
}

struct fsdir* fsadddirectory(struct fsdir* dir, char const* name, char const* keyname, struct idxalloc* ial) {
    struct varstrhashtable* vsh = dir->filenames;
    unsigned int eidx1 = vsaddstr(vsh, keyname, strlen(keyname), 0);
    struct fsdir* newdir = NULL;
    unsigned char result = 0;
    if (eidx1 != (unsigned int)-1) {
      result = 1;
      unsigned int fidx = fsgetnewdir(dir);
      if (fidx != (unsigned int)-1) {
        newdir = &dir->dirs[fidx];
        if (newdir->dirs == NULL) {
          result = fsinitdir(newdir, 4, 16, ial);
        } else {
          vsh->entries[eidx1 * 4 + 2] |= (fidx << 1 | 0) << 16;
          unsigned int eidx2 = vsaddstr(vsh, name, strlen(name), 0);
          newdir->filedata.parentdir = dir;
          newdir->filedata.keyname = eidx1;
          newdir->filedata.filename = eidx2 != (unsigned int)-1 ? eidx2 : eidx1;
          newdir->filedata.fileidx = idxgetnew(ial) << 1 | 0;
          if (eidx2 != (unsigned int)-1) {
            vsh->entries[eidx2 * 4 + 2] |= (fidx << 1 | 0) << 16;
          }
        }
      } else {
        result = 0;
      }
      if (result) {
        ++dir->dirnum;
      } else {
        newdir = NULL;
      }
    }
    return newdir;
}

void fsfreedir(struct fsdir* dir) {
  vshashtablefree(dir->filenames);
  free(dir->dirs);
  free(dir->files);
  dir->dirs = NULL;
  dir->files = NULL;
}

void fsfastremovedir(struct fsdir* dir, struct idxalloc* ial) {
  for (unsigned int fidx = 0; fidx < dir->filecap; ++fidx) {
    struct fsfile* file2 = &dir->files[fidx];
    if (file2->keyname != (unsigned int)-1) {
      idxremove(ial, file2->fileidx >> 1);
    }
  }
  for (unsigned int didx = 0; didx < dir->dircap; ++didx) {
    struct fsdir* subdir = &dir->dirs[didx];
    if (subdir->filedata.keyname != (unsigned int)-1) {
      fsfastremovedir(subdir, ial);
    }
  }
  idxremove(ial, dir->filedata.fileidx >> 1);
  fsfreedir(dir);
}

// remove file from parent, return id
// the file pointer should not be reused after this call
unsigned int fsremovefile(struct fsfile* file, struct idxalloc* ial) {
  struct fsdir* pdir = file->parentdir;
  if (file->keyname != (unsigned int)-1) {
    file->keyname = (unsigned int)-1; // mark as unallocated
    if (pdir) {
      struct varstrhashtable* vsh = pdir->filenames;
      vsremovebyidx(vsh, file->keyname);
      if (file->keyname != file->filename) {
        vsremovebyidx(vsh, file->filename);
      }
      if ((file->fileidx & 1)) {
        --pdir->filenum;
      } else {
        --pdir->dirnum;
      }
    }
  }
  idxremove(ial, file->fileidx >> 1);
  if ((file->fileidx & 1) == 0) { // remove subdirs/files
    struct fsdir* asdir = (struct fsdir*)((void*)file - offsetof(struct fsdir, filedata));
    fsfastremovedir(asdir, ial);
  }
  return file->fileidx;
}

char const* fsgetfilename(struct fsfile const* file) {
  char const* ret = NULL;
  if (file->parentdir != NULL) {
    struct varstrhashtable const* vsh = file->parentdir->filenames;
    ret = vsh->strdata + vsh->entries[file->filename * 4 + 1];
  }
  return ret;
}

char const* fsgetkeyname(struct fsfile const* file) {
  char const* ret = NULL;
  if (file->parentdir != NULL) {
    struct varstrhashtable const* vsh = file->parentdir->filenames;
    ret = vsh->strdata + vsh->entries[file->keyname * 4 + 1];
  }
  return ret;
}

void fsgetfilepath(struct fsfile const* file, char* buffer) {
  unsigned int pathlen = 0;
  struct fsdir const* currentparent = file->parentdir;
  struct fsfile const* currentfile = file;
  while (currentparent != NULL) {
    pathlen += currentparent->filenames->entries[currentfile->filename * 4 + 2] + 1;
    currentfile = &currentparent->filedata;
    currentparent = currentparent->filedata.parentdir;
  }
  unsigned int pathofs = pathlen;
  if ((file->fileidx & 1)) { // if file remove trailing /
    --pathlen;
  }
  currentparent = file->parentdir;
  currentfile = file;
  while (currentparent != NULL) {
    struct varstrhashtable const* vsh = currentparent->filenames;
    unsigned int slen = vsh->entries[currentfile->filename * 4 + 2];
    buffer[pathofs - 1] = '/';
    pathofs -= slen + 1;
    memcpy(buffer + pathofs, vsh->strdata + vsh->entries[currentfile->filename * 4 + 1], slen);
    currentfile = &currentparent->filedata;
    currentparent = currentparent->filedata.parentdir;
  }
  buffer[pathlen] = 0;
}

struct fsfile* fsgetfile(struct fsdir* dir, char const* keyname) {
  struct varstrhashtable const* vsh = dir->filenames;
  unsigned int eidx = vsfindstr(vsh, keyname, strlen(keyname));
  struct fsfile* rfile = NULL;
  if(eidx != (unsigned int)-1) {
    unsigned int vidx = (vsh->entries[eidx * 4 + 2] >> 16);
    rfile = (vidx & 1) ? &dir->files[vidx >> 1] : &dir->dirs[vidx >> 1].filedata;
  }
  return rfile;
}

struct fsfile* fsresolvepath(struct fsdir* dir, char const* path) {
  unsigned int pathofs = 0;
  unsigned int lastpathofs = 0;
  unsigned int const pathlen = strlen(path);
  struct fsfile* cfile = &dir->filedata;
  while (pathofs <= pathlen && cfile != NULL && !(cfile->fileidx & 1)) {
    if (path[pathofs] == '/' || path[pathofs] == 0) {
      if (pathofs > lastpathofs) {
        struct fsdir* asdir = (struct fsdir*)((void*)cfile - offsetof(struct fsdir, filedata));
        unsigned int eidx = vsfindstr(asdir->filenames, path + lastpathofs, pathofs - lastpathofs);
        if (eidx != (unsigned int)-1) {
          unsigned int vidx = (asdir->filenames->entries[eidx * 4 + 2] >> 16);
          cfile = (vidx & 1) ? &asdir->files[vidx >> 1] : &asdir->dirs[vidx >> 1].filedata;
        } else {
          cfile = NULL;
        }
      }
      lastpathofs = pathofs + 1;
    }
    ++pathofs;
  }
  return pathofs <= pathlen ? NULL : cfile;
}
